==> prism-2.1.5-20250625.094403-10.jar: prism/logging.clj <==
(ns prism.logging
  (:require
    [clojure.stacktrace :as stacktrace]
    [clojure.string :as str]
    [prism.core :as prism]
    [prism.internal.classpath :as cp]
    [prism.json :as json]
    [taoensso.encore]
    [taoensso.telemere :as tel]
    [taoensso.telemere.timbre :as telemere-timbre])
  (:import
    (java.io Writer)))

(defn structure-timbre-log [{:keys [msg_ ?err ?ns-str ?line level timestamp_ context]}]
  (let [thread (Thread/currentThread)
        thread-str (or (not-empty (.getName thread))
                       (str "virtual-thread-" (.threadId thread)))]
    (cond-> (assoc context
              :caller (str ?ns-str \: ?line)
              :level (name level)
              :thread thread-str
              :msg (force msg_)
              :ts (force timestamp_))
            ?err (assoc :stacktrace (with-out-str
                                      (stacktrace/print-cause-trace ?err)))
            (seq (ex-data ?err)) (assoc :ex-data (ex-data ?err)))))

(def configure-timbre-logging! (cp/missing-dep "jsonista" "aero" "timbre"))
(cp/when-ns 'taoensso.timbre
  (cp/when-ns 'prism.json
    (cp/when-ns 'aero.core
      (defn configure-logging! []
        (cp/if-ns 'taoensso.telemere.timbre
          (taoensso.timbre/set-config!
            {:min-level      (-> (prism/config) :log-level keyword)
             :ns-filter      #{"*"}
             :middleware     [(fn structure-log-middleware [m]
                                {:output_ (-> (structure-timbre-log m)
                                              prism.json/write-json-string
                                              delay)})]
             :timestamp-opts taoensso.timbre/default-timestamp-opts
             :appenders      {:println (telemere-timbre/timbre->telemere-appender)}})
          (taoensso.timbre/set-config!
            {:min-level      (-> (prism/config) :log-level keyword)
             :ns-filter      #{"*"}
             :middleware     []
             :timestamp-opts taoensso.timbre/default-timestamp-opts
             :output-fn      structure-timbre-log
             :appenders      {:println {:enabled?   true
                                        :async?     false
                                        :min-level  nil
                                        :rate-limit nil
                                        :output-fn  :inherit
                                        :fn         (fn json-appender-fn [{:keys [output_]}]
                                                      (.write ^Writer *err*
                                                              (str
                                                                (prism.json/write-json-string (force output_))
                                                                taoensso.encore/system-newline))
                                                      (.flush ^Writer *err*))}}}))))))


(defn structure-telemere-log [{:keys [inst ns coords kind level id uid msg_ data error ctx thread]}]
  (when (= kind :log)
    (let [ex-data (ex-data error)]
      (cond-> (assoc data
                :caller (str ns \: (str/join \: coords))
                :level (name level)
                :ts inst
                :thread (or (:name thread)
                            (str (:group thread "virtual-thread")
                                 "-"
                                 (:id thread)))
                :msg (force msg_)
                :id id
                :uid uid)
              ctx (assoc :context ctx)
              error (assoc :stacktrace (with-out-str
                                         (stacktrace/print-cause-trace error)))
              ex-data (assoc :ex-data ex-data)))))

(defn configure-telemere-logging! []
  (-> (prism/config) :log-level keyword tel/set-min-level!)
  (cp/when-ns 'prism.json
    (tel/add-handler!
      :default/console
      (tel/handler:console {:stream    :err
                            :output-fn (tel/pr-signal-fn {:clean-fn identity
                                                          :pr-fn    (fn pr-log [signal]
                                                                      (some-> (structure-telemere-log signal)
                                                                              json/write-json-string))})}))))

(comment
  (configure-telemere-logging!)
  (tel/log! {:level :warn
             :data  {:m 1}
             :id    ::x}
            "test")
  (tel/log! {:level :debug
             :id    ::y}
            "test"))

==> prism-2.1.5-20250625.095547-11.jar: prism/logging.clj <==
(ns prism.logging
  (:require
    [clojure.stacktrace :as stacktrace]
    [clojure.string :as str]
    [prism.core :as prism]
    [prism.internal.classpath :as cp]
    [prism.json :as json]
    [taoensso.encore]
    [taoensso.telemere :as tel]
    [taoensso.telemere.timbre :as telemere-timbre])
  (:import
    (java.io Writer)))

(defn structure-timbre-log [{:keys [msg_ ?err ?ns-str ?line level timestamp_ context]}]
  (let [thread (Thread/currentThread)
        thread-str (or (not-empty (.getName thread))
                       (str "virtual-thread-" (.threadId thread)))]
    (cond-> (assoc context
              :caller (str ?ns-str \: ?line)
              :level (name level)
              :thread thread-str
              :msg (force msg_)
              :ts (force timestamp_))
            ?err (assoc :stacktrace (with-out-str
                                      (stacktrace/print-cause-trace ?err)))
            (seq (ex-data ?err)) (assoc :ex-data (ex-data ?err)))))

(def configure-timbre-logging! (cp/missing-dep "jsonista" "aero" "timbre"))
(cp/when-ns 'taoensso.timbre
  (cp/when-ns 'prism.json
    (cp/when-ns 'aero.core
      (defn configure-logging! []
        (cp/if-ns 'taoensso.telemere.timbre
          (taoensso.timbre/set-config!
            {:min-level      (-> (prism/config) :log-level keyword)
             :ns-filter      #{"*"}
             :middleware     [(fn structure-log-middleware [m]
                                {:output_ (-> (structure-timbre-log m)
                                              prism.json/write-json-string
                                              delay)})]
             :timestamp-opts taoensso.timbre/default-timestamp-opts
             :appenders      {:println (telemere-timbre/timbre->telemere-appender)}})
          (taoensso.timbre/set-config!
            {:min-level      (-> (prism/config) :log-level keyword)
             :ns-filter      #{"*"}
             :middleware     []
             :timestamp-opts taoensso.timbre/default-timestamp-opts
             :output-fn      structure-timbre-log
             :appenders      {:println {:enabled?   true
                                        :async?     false
                                        :min-level  nil
                                        :rate-limit nil
                                        :output-fn  :inherit
                                        :fn         (fn json-appender-fn [{:keys [output_]}]
                                                      (.write ^Writer *err*
                                                              (str
                                                                (prism.json/write-json-string (force output_))
                                                                taoensso.encore/system-newline))
                                                      (.flush ^Writer *err*))}}}))))))


(defn structure-telemere-log [{:keys [inst ns coords kind level id uid msg_ data error ctx thread]}]
  (when (= kind :log)
    (let [ex-data (ex-data error)]
      (cond-> (assoc data
                :caller (str ns \: (str/join \: coords))
                :level (name level)
                :ts inst
                :thread (or (:name thread)
                            (str (:group thread "virtual-thread")
                                 "-"
                                 (:id thread)))
                :msg (force msg_)
                :id id)
              uid (assoc :uid uid)
              ctx (assoc :context ctx)
              error (assoc :stacktrace (with-out-str
                                         (stacktrace/print-cause-trace error)))
              ex-data (assoc :ex-data ex-data)))))

(defn configure-telemere-logging! []
  (-> (prism/config) :log-level keyword tel/set-min-level!)
  (tel/remove-handler! :default/console)
  (cp/when-ns 'prism.json
    (tel/add-handler!
      ::json-console-handler
      (tel/handler:console
        {:stream    :err
         (tel/pr-signal-fn {:clean-fn identity
                                       :pr-fn    (fn pr-log [signal]
                                                   (some-> (structure-telemere-log signal)
                                                           json/write-json-string))})})
      {:track-stats? false})))

(comment
  (configure-telemere-logging!)
  (tel/log! {:level :warn
             :data  {:m 1}
             :id    ::x}
            "test")
  (tel/log! {:level :debug
             :id    ::y}
            "test"))
